/*$Id: s_ac.cc,v 26.81 2008/05/27 05:34:00 al Exp $ -*- C++ -*-
 * Copyright (C) 2001 Albert Davis
 * Author: Albert Davis <aldavis@gnu.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 * ac analysis top
 */
//testing=script 2006.07.14
#include "globals.h"
#include "u_prblst.h"
#include "s__.h"
/*--------------------------------------------------------------------------*/
namespace {
/*--------------------------------------------------------------------------*/
class AC : public SIM {
public:
  void	do_it(CS&, CARD_LIST* scope);

  explicit AC():
    SIM(),
    _start(0.),
    _stop(0.),
    _step(0.),
    _linswp(false),
    _prevopppoint(false)
  {}

  ~AC() {}
private:
  explicit AC(const AC&):SIM() {unreachable(); incomplete();}
  void	sweep();
  void	first();
  bool	next();
  void	solve();
  void	clear();
  void	setup(CS&);
  bool	by(CS&);
  bool	decade(CS&);
  bool	lin(CS&);
  bool	octave(CS&);
  bool	times(CS&);
private:
  double _start;	// sweep start frequency
  double _stop;		// sweep stop frequency
  double _step;		// printed step size
  bool _linswp;		// flag: use linear sweep (vs log sweep)
  bool _prevopppoint;	// flag: use previous op point
};
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void AC::do_it(CS& cmd, CARD_LIST*)
{
  set_command_ac();
  reset_timers();
  ::status.ac.reset().start();

  init();
  alloc_vectors();
  
  acx.reallocate();
  acx.set_min_pivot(OPT::pivtol);
  
  setup(cmd);
  ::status.set_up.stop();
  switch (ENV::run_mode) {
  case rPRE_MAIN:	unreachable();	break;
  case rBATCH:		untested();
  case rINTERACTIVE:    itested();
  case rSCRIPT:		sweep();	break;
  case rPRESET:		/*nothing*/	break;
  }
  acx.unallocate();
  unalloc_vectors();
  
  ::status.ac.stop();
  ::status.total.stop();
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
static int needslinfix;	// flag: lin option needs patch later (spice compat)
/*--------------------------------------------------------------------------*/
void AC::setup(CS& cmd)
{
  out = IO::mstdout;
  out.reset(); //BUG// don't know why this is needed
  temp_c = OPT::temp_c;
  bool ploton = IO::plotset  &&  plotlist().size() > 0;

       if (cmd.umatch("*"))	    {untested(); times(cmd);}
  else if (cmd.umatch("by "))       {untested(); by(cmd);}
  else if (cmd.umatch("step "))     {untested(); by(cmd);}
  else if (cmd.umatch("d{ecade} ")) {		 decade(cmd);}
  else if (cmd.umatch("ti{mes} "))  {untested(); times(cmd);}
  else if (cmd.umatch("lin "))	    {untested(); lin(cmd);}
  else if (cmd.umatch("o{ctave} ")) {		 octave(cmd);}
  
  if (cmd.is_float()) {
    _start = cmd.ctof();
    _stop  = cmd.ctof();
    if (_stop==0.) {
      _stop = _start;
    }
    if (cmd.is_float()) {
      by(cmd);
    }
  }

  unsigned here = cmd.cursor();
  do{
    ONE_OF
      || (cmd.umatch("*")	  && times(cmd))
      || (cmd.umatch("+")	  && by(cmd))
      || (cmd.umatch("by ")	  && by(cmd))
      || (cmd.umatch("step ")	  && by(cmd))
      || (cmd.umatch("d{ecade} ") && decade(cmd))
      || (cmd.umatch("ti{mes} ")  && times(cmd))
      || (cmd.umatch("lin ")	  && lin(cmd))
      || (cmd.umatch("o{ctave} ") && octave(cmd))
      || Get(cmd, "dt{emp}",	   &temp_c,  mOFFSET, OPT::temp_c)
      || Get(cmd, "pl{ot}",	   &ploton)
      || Get(cmd, "pr{evoppoint}", &_prevopppoint)
      || Get(cmd, "sta{rt}",	   &_start)
      || Get(cmd, "sto{p}",	   &_stop)
      || Get(cmd, "te{mperature}", &temp_c)
      || outset(cmd,&out)
      ;
  }while (cmd.more() && !cmd.stuck(&here));
  cmd.check(bWARNING, "what's this?");

  IO::plotout = (ploton) ? IO::mstdout : OMSTREAM();
  initio(out);
  if (needslinfix) {untested();		// LIN option is # of points.
    assert(_step >= 2);			// Must compute step after 
    _step=(_stop-_start)/(_step-1.);	// reading start and stop,
    needslinfix = false;		// but step must be read first
  }					// for Spice compatibility
  if (_step==0.) {
    _step = _stop - _start;
    _linswp = true;
  }
}
/*--------------------------------------------------------------------------*/
bool AC::by(CS& cmd)
{
  _step = cmd.ctof();
  needslinfix = false;
  _linswp = true;
  return true;
}
/*--------------------------------------------------------------------------*/
bool AC::decade(CS& cmd)
{
  _step = cmd.ctopf();
  if (_step == 0.) {
    _step = 1.;
  }
  _step = pow(10., 1./_step);
  needslinfix = false;
  _linswp = false;
  return true;
}
/*--------------------------------------------------------------------------*/
bool AC::lin(CS& cmd)
{untested();
  _step = cmd.ctopf();		// need to fix step, later
  if (_step <= 2.) {untested();	// do it at the end of setup
    _step = 2.;			// a kluge, but this is a patch
  }else{untested();
  }
  needslinfix = true;		// and I am too lazy to do it
  _linswp = true;		// right.
  return true;
}
/*--------------------------------------------------------------------------*/
bool AC::octave(CS& cmd)
{
  _step = cmd.ctopf();
  if (_step == 0.) {untested();
    _step = 1.;
  }
  _step = pow(2.00000001, 1./_step);
  needslinfix = false;
  _linswp = false;
  return true;
}
/*--------------------------------------------------------------------------*/
bool AC::times(CS& cmd)
{untested();
  _step = cmd.ctopf();
  if (_step == 0.  &&  _start != 0.) {untested();
    _step = _stop / _start;
    untested();
  }else{untested();
  }
  needslinfix = false;
  _linswp = false;
  return true;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void AC::solve(void)
{
  acx.zero();
  std::fill_n(ac, ::status.total_nodes+1, 0.);

  ::status.load.start();
  count_iterations(iTOTAL);
  CARD_LIST::card_list.do_ac();
  ::status.load.stop();

  ::status.lud.start();
  acx.lu_decomp();
  ::status.lud.stop();

  ::status.back.start();
  acx.fbsub(ac);
  ::status.back.stop();
}
/*--------------------------------------------------------------------------*/
void AC::sweep()
{
  head(_start, _stop, "Freq");
  first();
  CARD_LIST::card_list.ac_begin();
  do {
    jomega = COMPLEX(0., freq * M_TWO_PI);
    solve();
    outdata(freq);
  } while (next());
}
/*--------------------------------------------------------------------------*/
void AC::first()
{
  freq = _start;
}
/*--------------------------------------------------------------------------*/
bool AC::next()
{
  double realstop = (_linswp)
    ? _stop - _step/100.
    : _stop / pow(_step,.01);
  if (!in_order(_start, freq, realstop)) {
    return false;
  }

  freq = (_linswp)
    ? freq + _step
    : freq * _step;
  if (in_order(freq, _start, _stop)) {
    return false;
  }else{
    return true;
  }
}
/*--------------------------------------------------------------------------*/
static AC p1;
static DISPATCHER<CMD>::INSTALL d1(&command_dispatcher, "ac", &p1);
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
