/*$Id: s_dc.cc,v 26.83 2008/06/05 04:46:59 al Exp $ -*- C++ -*-
 * Copyright (C) 2001 Albert Davis
 * Author: Albert Davis <aldavis@gnu.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 * dc analysis top
 */
//testing=script,complete 2006.07.14
#include "u_prblst.h"
#include "u_cardst.h"
#include "e_elemnt.h"
#include "globals.h"
/*--------------------------------------------------------------------------*/
namespace {
/*--------------------------------------------------------------------------*/
class DCOP : public SIM {
public:				/* s_dc_set.cc */
  void	finish();
protected:
  void	options(CS&, int);
private:
  bool	by(CS&, int);
  bool	decade(CS&, int);
  bool	times(CS&, int);
private:			/* s_dc_swp.cc */
  void	sweep();
  void	sweep_recursive(int);
  void	first(int);
  bool	next(int);
  explicit DCOP(const DCOP&): SIM() {unreachable(); incomplete();}
protected:
  explicit DCOP():
    SIM(),
    n_sweeps(1),
    loop(false),
    reverse(false),
    cont(false),
    trace(tNONE)
  {
    for (int ii = 0; ii < DCNEST; ++ii) {
      start[ii]=stop[ii]=step[ii]=0.;
      linswp[ii]=true;
      sweepval[ii]=&genout;
      zap[ii]=NULL; 
    }

    //BUG// in SIM.  should be initialized there.
    genout=0.;
    temp_c=OPT::temp_c;
    out=IO::mstdout;
    uic=false;
  }
  ~DCOP() {}
  
protected:
  enum {DCNEST = 4};
  int n_sweeps;
  double start[DCNEST];
  double stop[DCNEST];
  double step[DCNEST];
  bool linswp[DCNEST];
  double* (sweepval[DCNEST]);	/* pointer to thing to sweep, dc command */
  ELEMENT* (zap[DCNEST]);	/* to branch to zap, for re-expand */
  CARDSTASH stash[DCNEST];	/* store std values of elements being swept */
private:
  bool loop;			/* flag: do it again backwards */
  bool reverse;			/* flag: sweep backwards */
  bool cont;			/* flag: continue from previous run */
  TRACE trace;			/* enum: show extended diagnostics */
};
/*--------------------------------------------------------------------------*/
class DC : public DCOP {
public:
  explicit DC(): DCOP() {}
  ~DC() {}
  void	do_it(CS&, CARD_LIST* scope);
private:
  void	setup(CS&);
  explicit DC(const DC&): DCOP() {unreachable(); incomplete();}
};
/*--------------------------------------------------------------------------*/
class OP : public DCOP {
public:
  explicit OP(): DCOP() {}
  ~OP() {}
  void	do_it(CS&, CARD_LIST* scope);
private:
  void	setup(CS&);
  explicit OP(const OP&): DCOP() {unreachable(); incomplete();}
};
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void DC::do_it(CS& cmd, CARD_LIST*)
{
  time0 = 0.;
  set_command_dc();
  _phase = p_INIT_DC;
  ::status.dc.reset().start();
  command_base(cmd);
  ::status.dc.stop();
}
/*--------------------------------------------------------------------------*/
void OP::do_it(CS& cmd, CARD_LIST*)
{
  time0 = 0.;
  set_command_op();
  _phase = p_INIT_DC;
  ::status.op.reset().start();
  command_base(cmd);
  ::status.op.stop();
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void DCOP::finish(void)
{
  for (int ii = 0;  ii < n_sweeps;  ++ii) {
    if (exists(zap[ii])) { // component
      stash[ii].restore();
      zap[ii]->dec_probes();
      zap[ii]->elaborate();
    }
  }
}
/*--------------------------------------------------------------------------*/
void OP::setup(CS& cmd)
{
  zap[0] = NULL;
  sweepval[0] = &temp_c;
  start[0] = (cmd.is_float()) ? cmd.ctof() : OPT::temp_c;
  stop[0]  = (cmd.is_float()) ? cmd.ctof() : start[0];
  step[0] = 0.;
  genout = 0.;
  options(cmd,0);
  if (step[0] == 0.) {
    step[0] = stop[0] - start[0];
    linswp[0] = true;
  }else{
    itested();
  }
  n_sweeps = 1;
  cmd.check(bWARNING, "what's this?");
  freq = 0;
}
/*--------------------------------------------------------------------------*/
void DC::setup(CS& cmd)
{
  if (cmd.more()) {
    for (n_sweeps = 0; cmd.more() && n_sweeps < DCNEST; ++n_sweeps) {
      CARD_LIST::fat_iterator ci = findbranch(cmd, &CARD_LIST::card_list);
      if (!ci.is_end()) {			// sweep a component
	if (ELEMENT* c = dynamic_cast<ELEMENT*>(*ci)) {
	  zap[n_sweeps] = c;
	}else{untested();
	  throw Exception("dc/op: can't sweep " + (**ci).long_label() + '\n');
	}
      }else if (cmd.is_float()) {		// sweep the generator
	zap[n_sweeps] = NULL;
      }else{
	// leave as it was .. repeat cmd with no args
      }
      
      if (cmd.is_float()) {			// set up parameters
	start[n_sweeps] = cmd.ctof();
	stop[n_sweeps] = (cmd.is_float()) ? cmd.ctof() : start[n_sweeps];
	step[n_sweeps] = 0.;
      }else{
	// leave it as it was .. repeat cmd with no args
      }
      
      genout = 0.;
      temp_c = OPT::temp_c;
      options(cmd,n_sweeps);
      if (step[n_sweeps] == 0.) {	// prohibit log sweep from 0
	step[n_sweeps] = stop[n_sweeps] - start[n_sweeps];
	linswp[n_sweeps] = true;
      }
    }
  }else{
    options(cmd,n_sweeps);	// there is important initialization here
  }
  cmd.check(bWARNING, "what's this?");

  assert(n_sweeps > 0);
  for (int ii = 0;  ii < n_sweeps;  ++ii) {
    if (exists(zap[ii])) { // component
      stash[ii] = zap[ii];			// stash the std value
      zap[ii]->inc_probes();			// we need to keep track of it
      zap[ii]->set_value(zap[ii]->value(),0);	// zap out extensions
      zap[ii]->set_constant(false);		// so it will be updated
      sweepval[ii] = zap[ii]->set__value();	// point to value to patch
    }else{ // generator
      sweepval[ii] = &genout;			// point to value to patch
    }
  }
  freq = 0;
}
/*--------------------------------------------------------------------------*/
void DCOP::options(CS& cmd, int nest)
{
  out = IO::mstdout;
  out.reset(); //BUG// don't know why this is needed */
  bool ploton = IO::plotset  &&  plotlist().size() > 0;
  uic = loop = reverse = cont = false;
  trace = tNONE;
  unsigned here = cmd.cursor();
  do{
    ONE_OF
      || (cmd.is_float()	 && by(cmd,nest))
      || (cmd.umatch("*")	 && times(cmd,nest))
      || (cmd.umatch("by ")	 && by(cmd,nest))
      || (cmd.umatch("de{cade} ")&& decade(cmd,nest))
      || (cmd.umatch("ti{mes} ") && times(cmd,nest))
      || Get(cmd, "c{ontinue}",   &cont)
      || Get(cmd, "dt{emp}",	  &temp_c,   mOFFSET, OPT::temp_c)
      || Get(cmd, "lo{op}", 	  &loop)
      || Get(cmd, "pl{ot}",	  &ploton)
      || Get(cmd, "re{verse}",	  &reverse)
      || Get(cmd, "te{mperature}",&temp_c)
      || (cmd.umatch("tr{ace} {=}") &&
	  (ONE_OF
	   || Set(cmd, "n{one}",      &trace, tNONE)
	   || Set(cmd, "o{ff}",       &trace, tNONE)
	   || Set(cmd, "w{arnings}",  &trace, tUNDER)
	   || Set(cmd, "i{terations}",&trace, tITERATION)
	   || Set(cmd, "v{erbose}",   &trace, tVERBOSE)
	   || cmd.warn(bWARNING, 
		       "need none, off, warnings, iterations, verbose")
	   )
	  )
      || outset(cmd,&out)
      ;
  }while (cmd.more() && !cmd.stuck(&here));

  IO::plotout = (ploton) ? IO::mstdout : OMSTREAM();
  initio(out);
}
/*--------------------------------------------------------------------------*/
bool DCOP::by(CS& cmd, int nest)
{
  step[nest] = cmd.ctof();
  linswp[nest] = true;
  if (step[nest] == 0.) {
    untested();
    step[nest] = stop[nest] - start[nest];
  }
  return true;
}
/*--------------------------------------------------------------------------*/
bool DCOP::decade(CS& cmd, int nest)
{
  double junk = cmd.ctopf();
  if (junk == 0.) {
    untested();
    junk = 1.;
  }
  junk = pow(10., 1./junk);
  step[nest] = junk;
  linswp[nest] = false;
  return true;
}
/*--------------------------------------------------------------------------*/
bool DCOP::times(CS& cmd, int nest)
{
  untested();
  step[nest] = cmd.ctopf();
  linswp[nest] = false;
  if (step[nest] == 0.  &&  start[nest] != 0.) {
    untested();
    step[nest] = stop[nest] / start[nest];
  }else{
    untested();
  }
  return true;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void DCOP::sweep(void)
{
  head(start[0], stop[0], " ");
  bypass_ok = false;
  set_inc_mode_bad();
  if (cont) {untested();
    restore_voltages();
  }
  
  clear_limit();
  CARD_LIST::card_list.tr_begin();
  sweep_recursive(n_sweeps);
}
/*--------------------------------------------------------------------------*/
void DCOP::sweep_recursive(int ii)
{
  --ii;
  assert(ii >= 0);
  assert(ii < DCNEST);

  OPT::ITL itl = OPT::DCBIAS;
  
  first(ii);
  do {
    if (ii == 0) {
      int converged = solve(itl,trace);
      if (!converged) {itested();
	error(bWARNING, "did not converge\n");
      }
      ::status.accept.start();
      set_limit();
      CARD_LIST::card_list.tr_accept();
      ::status.accept.stop();
      keep_voltages();
      outdata(*sweepval[ii]);
      itl = OPT::DCXFER;
    }else{
      sweep_recursive(ii);
    }
  } while (next(ii));
}
/*--------------------------------------------------------------------------*/
void DCOP::first(int ii)
{
  assert(ii >= 0);
  assert(ii < DCNEST);
  assert(start);
  assert(sweepval);
  assert(sweepval[ii]);

  *sweepval[ii] = start[ii];
  if (reverse) {itested();
    reverse = false;
    while (next(ii)) {itested();
      /* nothing */;
    }
    reverse = true;
    next(ii);
  }
  _phase = p_INIT_DC;
}
/*--------------------------------------------------------------------------*/
bool DCOP::next(int ii)
{
  bool ok = false;
  if (linswp[ii]) {
    double fudge = step[ii] / 10.;
    if (step[ii] == 0.) {
      ok = false;
    }else{
      if (!reverse) {
	*(sweepval[ii]) += step[ii];
	fixzero(sweepval[ii], step[ii]);
	ok=in_order(start[ii]-fudge,*(sweepval[ii]),stop[ii]+fudge);
	if (!ok  &&  loop) {
	  reverse = true;
	}
      }
      if (reverse) {
	*(sweepval[ii]) -= step[ii];
	fixzero(sweepval[ii], step[ii]);
	ok=in_order(start[ii]-fudge,*(sweepval[ii]),stop[ii]+fudge);
      }
    }
  }else{
    double fudge = pow(step[ii], .1);
    if (step[ii] == 1.) {untested();
      ok = false;
    }else{
      if (!reverse) {
	*(sweepval[ii]) *= step[ii];
	ok=in_order(start[ii]/fudge,*(sweepval[ii]),stop[ii]*fudge);
	if (!ok  &&  loop) {untested();
	  reverse = true;
	}     
      }
      if (reverse) {untested();
	*(sweepval[ii]) /= step[ii];
	ok=in_order(start[ii]/fudge,*(sweepval[ii]),stop[ii]*fudge);
      }
    }
  }
  _phase = p_DC_SWEEP;
  return ok;
}
/*--------------------------------------------------------------------------*/
static DC p2;
static OP p4;
static DISPATCHER<CMD>::INSTALL d2(&command_dispatcher, "dc", &p2);
static DISPATCHER<CMD>::INSTALL d4(&command_dispatcher, "op", &p4);
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
