/* $Id: spice-wrapper.cc,v 26.83 2008/06/05 04:46:59 al Exp $ -*- C++ -*-
 * Copyright (C) 2007 Albert Davis
 * Author: Albert Davis <aldavis@gnu.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This file is distributed as is, completely without warranty or
 * service support. The author and its employees are not liable for
 * the condition or performance of the software.
 *
 * The author owns the copyright but shall not be liable for any
 * infringement of copyright or other proprietary rights brought by
 * third parties against the users of the software.
 *
 * The author hereby disclaims all implied warranties.
 *
 * This author grants the users the right to modify, copy, and
 * redistribute this file, for any purpose, both within the user's
 * organization and externally.
 */
//testing=informal
// code style comment:  Use of "reinterpret_cast" is always bad style.
/*--------------------------------------------------------------------------*/
// spice includes
extern "C" {
  #define _complex CompleX
  #define NODE NodE
  #define public PubliC
  #define bool BooL
  #define main MaiN
  #include "capabil.h"
  #include "const.h"
  #include "iferrmsg.h"
  #include "devdefs.h"
  #include "ftedefs.h"
  #include "optdefs.h"
#ifdef JSPICE3
  #include "uflags.h"
  #include "inpdefs.h"
#endif
  #undef main
  #undef bool
  #undef public
  #undef NODE
  #undef _complex

  #undef eq
  #undef OPT
  #undef LINEAR
  #undef STRING
  #undef BOOLEAN
  
  #undef VT_BOOL
  #undef VT_NUM
  #undef VT_REAL
  #undef VT_STRING
  #undef VT_LIST
}
/*--------------------------------------------------------------------------*/
// gnucap includes
#include "globals.h"
#include "e_storag.h"
#include "e_model.h"
/*--------------------------------------------------------------------------*/
// customization -- must be last
#include "wrapper.h"
#if !defined(UNCONNECTED_NODES)
  #define UNCONNECTED_NODES DISALLOW
  #if (MIN_NET_NODES != MAX_NET_NODES)
    #error "What should I do with the unconnected nodes?"
  #endif
#endif
/*--------------------------------------------------------------------------*/
extern SPICEdev info;
const int SPICE_INVALID_NODE = 0;
const int SPICE_UNCONNECTED_NODE = -1;
const int OFFSET = 1;
#define GROUND 1
#define FLOAT 2
#define DISALLOW 3
class DEV_SPICE;
class MODEL_SPICE;
/*--------------------------------------------------------------------------*/
/* function mapping: see devdefs.h
 * DEVparam	DEV_SPICE::parse_spice
 * DEVmodParam	MODEL_SPICE::parse_params
 * DEVload	DEV_SPICE::do_tr
 * DEVsetup	MODEL_SPICE::elaborate, DEV_SPICE::elaborate
 * DEVunsetup	not used -- spice baggage -- just zeros some nodes
 * DEVpzSetup	not used -- pole-zero
 * DEVtemperature DEV_SPICE::precalc
 * DEVtrunc	DEV_SPICE::tr_review
 * DEVfindBranch not used -- current probes for current controlled source
 * DEVacLoad	DEV_SPICE::do_ac
 * DEVaccept 	not used -- sets break points //BUG// need this
 * DEVdestroy	not used -- spice baggage -- deletes a list
 * DEVmodDelete	not used -- spice baggage -- delete one model
 * DEVdelete	not used -- spice baggage -- delete one instance
 * DEVsetic  	not used -- "getic" -- initial conditions //BUG// need this
 * DEVask	DEV_SPICE::print_args, DEV_SPICE::tr_probe_num
 * DEVmodAsk	MODEL_SPICE::print_params, MODEL_SPICE::print_calculated
 * DEVpzLoad  	not used -- pole zero
 * DEVconvTest 	DEV_SPICE::do_tr
 * DEVsenSetup 	not used -- sensitivity
 * DEVsenLoad  	not used -- sensitivity
 * DEVsenUpdate	not used -- sensitivity
 * DEVsenAcLoad	not used -- sensitivity
 * DEVsenPrint 	not used -- sensitivity
 * DEVsenTrunc 	not used -- sensitivity
 * DEVdisto    	not used -- distorion
 * DEVnoise   	not used -- noise
*/
/*--------------------------------------------------------------------------*/
union MODEL_PTR {
  GENmodel* _g; // gendefs.h
  MODEL*    _m;
  char*	    _c;

  bool assert_invariant()const {
    assert(_g);
    assert(_g->GENmodType == 0);
    assert(!(_g->GENnextModel));
  }
  bool assert_raw()const {
    assert_invariant();
    assert(!(_g->GENinstances));
  }
  bool assert_setup()const {
    assert_invariant();
    assert(_g->GENinstances);
    assert(_g->GENmodName);
  }

  MODEL_PTR(GENmodel* G) : _g(G) {}
  MODEL_PTR(MODEL* M) : _m(M) {}
  MODEL_PTR() : _g(NULL)    {assert(!_m);}
  operator bool()const	    {return _g;}
  operator bool()	    {return _g;}
  operator char*()	    {return _c;}
  operator GENmodel*()const {return _g;}
  operator MODEL*()	    {return _m;}

  // see spice include gendefs.h
  int		GENmodType()const	{return _g->GENmodType;}
  GENmodel*	GENnextModel()const	{return _g->GENnextModel;}
  GENinstance*&	GENinstances()const	{return _g->GENinstances;}
  IFuid&	GENmodName()		{return _g->GENmodName;}

  void set_mod_name(std::string s) {
    char* p = new char[s.length()+1]; //BUG//memory leak
    s.copy(p, std::string::npos);
    p[s.length()] = '\0';
    _g->GENmodName = p;
  }
};
/*--------------------------------------------------------------------------*/
union MATRIX_PTR {
  SMPmatrix* _s;
  COMPLEX*   _m;
};
/*--------------------------------------------------------------------------*/
class MODEL_SPICE
  :public MODEL_CARD{
private:
  static int _count;
  static CKTcircuit _ckt;
public:
  union {
    mutable GENmodel _gen_model_raw;
    MODEL _model_raw;
    char  _model_raw_space;
  };
  MODEL_PTR _model_setup;
  std::string _key;
  std::string _level;
protected:
  explicit MODEL_SPICE(const MODEL_SPICE& p);
public:
  explicit MODEL_SPICE(const DEV_SPICE* p);
  ~MODEL_SPICE();
public: // override virtual
  MODEL_CARD* clone()const {return new MODEL_SPICE(*this);}
  void elaborate();

  //--------------------------------------------------------------------
public:	// type
  void set_dev_type(const std::string& nt);
  std::string dev_type()const	{return _key;}
  //--------------------------------------------------------------------
public: // parameters
  //bool param_exists(int)const;
  bool param_is_printable(int)const;
  std::string param_name(int)const;
  std::string param_name(int i, int j)const {return MODEL_CARD::param_name(i, j);}
  std::string param_value(int)const; 
  //void set_param_by_name(std::string, std::string);
  void set_param_by_index(int, std::string&, int);
  //--------------------------------------------------------------------
public: // not virtual
  static int count()		{return _count;}
  static CKTcircuit* ckt()	{return &_ckt;}
  static void init_ckt();
};
/*--------------------------------------------------------------------------*/
#define MATRIX_NODES (MAX_NET_NODES + INTERNAL_NODES)

class DEV_SPICE : public STORAGE {
private:
  static int _count;
public:
private:
  union {
    mutable GENinstance _geninst;
    INSTANCE	_inst;
    char _inst_space;
  };
  std::string _modelname;
  MODEL_PTR _model_setup;
  node_t _nodes[MATRIX_NODES];
  COMPLEX* _matrix[MATRIX_NODES+OFFSET];	// For tran, real is now, imag is saved.
  COMPLEX _matrix_core[MATRIX_NODES+OFFSET][MATRIX_NODES+OFFSET];
public:
  double _i0[MATRIX_NODES+OFFSET];	// right side - current offsets
  double _i1[MATRIX_NODES+OFFSET];	// right side - saved
  double _v1[MATRIX_NODES+OFFSET];	// input voltages
  double* _state0;
  double* _state1;
  double* _state2;
  double* _state3;
  double* _state4;
  int _num_states;
private:
  explicit DEV_SPICE(const DEV_SPICE& p);
public:
  explicit DEV_SPICE();
  ~DEV_SPICE();
protected: // override virtual
  char	  id_letter()const	{return SPICE_LETTER[0];}
  bool	  print_type_in_spice()const {return true;}
  std::string value_name()const {return "#";}
  int	  max_nodes()const	{return MAX_NET_NODES;}
  int	  min_nodes()const	{return MIN_NET_NODES;}
  int	  matrix_nodes()const	{return MATRIX_NODES;}
  int	  net_nodes()const	{return _net_nodes;}
  int	  int_nodes()const	{return INTERNAL_NODES;}
  CARD*	  clone()const		{return new DEV_SPICE(*this);}
  void	  elaborate();
  //void  map_nodes();		//ELEMENT
  void	  precalc();

  void	  tr_iwant_matrix()	{tr_iwant_matrix_extended();}
  void	  tr_begin()		{STORAGE::tr_begin(); precalc();}
  void	  tr_restore();
  void	  dc_advance()		{STORAGE::dc_advance(); precalc();}
  void	  tr_advance();
  void	  tr_regress();
  bool	  tr_needs_eval()const	{return true;} //BUG//
  //void  tr_queue_eval();	//ELEMENT
  bool	  do_tr();
  void	  tr_load();
  TIME_PAIR tr_review();
  //void  tr_accept();		//CARD/nothing
  void	  tr_unload();
  double  tr_involts()const	{return NOT_VALID;}
  //double tr_input()const	//ELEMENT
  double  tr_involts_limited()const {return NOT_VALID;}
  //double tr_input_limited()const //ELEMENT
  //double tr_amps()const;	//ELEMENT
  double  tr_probe_num(const std::string&)const;

  void	  ac_iwant_matrix()	{ac_iwant_matrix_extended();}
  void    ac_begin()		{STORAGE::ac_begin(); precalc();}
  void	  do_ac();
  void	  ac_load();
  COMPLEX ac_involts()const	{return NOT_VALID;}
  COMPLEX ac_amps()const	{return NOT_VALID;}
  //XPROBE ac_probe_ext(CS&)const;//ELEMENT //BUG//
  int	  tail_size()const	{return 1;}
  //--------------------------------------------------------------------
public:	// type
  void set_dev_type(const std::string& nt);
  std::string dev_type()const	{return _modelname;}
  //--------------------------------------------------------------------
public:	// ports
  // bool port_exists(int i)const //COMPONENT
  std::string port_name(int i)const {itested();
    assert(i >= 0);
    assert(i < MAX_NET_NODES);
    return port_names[i];
  }
  // const std::string& port_value(int i)const; //COMPONENT
  //void set_port_by_name(std::string& name, std::string& value);
  //void set_port_by_index(int index, std::string& value);
  //--------------------------------------------------------------------
public: // parameters
  // bool param_exists(int)const;
  bool param_is_printable(int)const;
  std::string param_name(int)const;
  std::string param_name(int i, int j)const {return STORAGE::param_name(i, j);}
  std::string param_value(int)const; 
  // void set_param_by_name(std::string, std::string);
  void set_param_by_index(int, std::string&, int);
  //--------------------------------------------------------------------
private:
  CKTcircuit* ckt()const	{return MODEL_SPICE::ckt();}
  void init_ckt()		{MODEL_SPICE::init_ckt();}
  void update_ckt()const;
  void localize_ckt();
  int* spice_nodes()		{return &(_geninst.GENnode1);}
};
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
CKTcircuit MODEL_SPICE::_ckt;

/* used as intended, matters: CKTstates, CKTtime, CKTdelta, CKTtemp, CKTnomTemp,
 * CKTrhs, CKTrhsOld, CKTirhs, CKTmode, CKTabstol, CKTreltol, CKTvoltTol,
 * CKTgmin, CKTnoncon, CKTomega, CKTnoncon
 *
 * used as intended, doesn't matter: CKTintegrateMethod
 * 
 * used almost as intended, matters, probably ok: CKTdeltaOld[1:often, 2:ltra&tra],
 * CKTag, CKTorder(possible sload future problem), CKTbypass, CKTfixLimit(mos1236),
 * CKTsenInfo, CKTtroubleElt
 *
 * used almost as intended, matters, probably wrong: CKTtimePoints(ltra needs it)
 *
 * used almost as intended, doesn't matter: CKTmaxOrder

 * misused: (not used by spice devs, use to pass gnucap stuff through)
 * CKTstat: the device (type DEV_SPICE)
 * CKTmaxEqNum: use as counter for internal nodes
 *
 * need to handle: CKTcurrentAnalysis(ask),
 * CKTbreak([iv]src), CKTfinalTime([iv]src), CKTstep([iv]src), 
 * CKTminBreak(tra), CKTtrtol(ltra), CKTmaxStep(ltra), 
 * CKTdefaultMosL, CKTdefaultMosW, CKTdefaultMosAD, CKTdefaultMosAS, 
 * CKTtimeIndex(ltra), CKTsizeIncr(ltra), CKTtryToCompact(ltra),
 * CKTbadMos3(mos3)
 *
 * used by unused functions: CKThead, CKTirhsOld, CKTrhsOp,
 * CKTcurJob, 
 *
 * not used:   CKTvt, CKTmatrix, CKTniState, CKTrhsSpare, CKTirhsSpare,
 * CKTsenRhs, CKTseniRhs, CKTnodes, CKTlastNode, CKTnumStates,
 * CKTdcMaxIter, CKTdcTrcvMaxIter, CKTtranMaxIter, CKTbreakSize, CKTbreaks,
 * CKTsaveDelta, CKTminBreak, CKTbreaks, CKTpivotAbsTol, CKTpivotRelTol,
 * CKTchgtol, CKTlteReltol, CKTlteAbstol, CKTdelmin, CKTinitTime, CKTdiagGmin,
 * CKTnumSrcSteps, CKTnumGminSteps, CKTisSetup, CKTdeltaList, CKTtimeListSize,
 * CKTkeepOpInfo, CKThadNodeset, CKTnoOpIter, CKTtroubleNode,
 */
#ifdef JSPICE3
#define assert_ckt_initialized(ckt) {				\
    assert(ckt);						\
    assert((ckt)->CKTnomTemp == OPT::tnom_c + CONSTCtoK);	\
    assert((ckt)->CKTcurrentAnalysis == 0);			\
    assert((ckt)->CKTbypass  == false);				\
    assert((ckt)->CKTsrcFact == 1.);				\
  }
#else
#define assert_ckt_initialized(ckt) {				\
    assert(ckt);						\
    assert((ckt)->CKTnomTemp == OPT::tnom_c + CONSTCtoK);	\
    assert((ckt)->CKTcurrentAnalysis == 0);			\
    assert((ckt)->CKTbypass  == false);				\
    assert((ckt)->CKTsrcFact == 1.);				\
    assert((ckt)->CKTsenInfo == NULL);				\
  }
#endif

void MODEL_SPICE::init_ckt()
{
  assert(ckt());
  ckt()->CKTtime = SIM::time0;
  ckt()->CKTtemp    = SIM::temp_c + CONSTCtoK;
  ckt()->CKTnomTemp = OPT::tnom_c + CONSTCtoK;
  ckt()->CKTcurrentAnalysis = 0; // wrong but safe
  ckt()->CKTbypass  = false; // manage this elsewhere
  ckt()->CKTsrcFact = 1.; // source stepping kluge
  ckt()->CKTfixLimit = 1; // limiting kluge 1 == off
#ifndef JSPICE3
  ckt()->CKTsenInfo = NULL; // used as flag to print sens info
#endif
  assert_ckt_initialized(ckt());
}

#if 0
    assert(SIM::analysis_is_tran_dynamic() == (((ckt)->CKTmode & (MODE | MODEDC)) == MODETRAN));
    assert(SIM::analysis_is_tran_static() == (((ckt)->CKTmode & (MODE | MODEDC)) == MODETRANOP));
    assert(SIM::analysis_is_ac() == (((ckt)->CKTmode & (MODE | MODEDC)) == MODEAC));
    assert(SIM::command_is_dc() == (((ckt)->CKTmode & (MODE | MODEDC)) == MODEDCTRANCURVE));
    assert(SIM::command_is_op() == (((ckt)->CKTmode & (MODE | MODEDC)) == MODEDCOP));
#endif

#define assert_ckt_up_to_date(ckt) {				\
    assert_ckt_initialized(ckt);				\
    assert((ckt)->CKTtemp == SIM::temp_c + CONSTCtoK);		\
    assert((ckt)->CKTabstol == OPT::abstol);			\
    assert((ckt)->CKTreltol == OPT::reltol);			\
    assert((ckt)->CKTvoltTol == OPT::vntol);			\
    assert((ckt)->CKTgmin == OPT::gmin);			\
  }

void DEV_SPICE::update_ckt()const
{
  assert_ckt_initialized(ckt());
  ckt()->CKTtime = SIM::time0; // unreliable
  ckt()->CKTdelta = SIM::time0 - SIM::time1; // unreliable
  ckt()->CKTtemp = SIM::temp_c + CONSTCtoK; // unreliable
  ckt()->CKTmode = 0;
  if (analysis_is_tran()) {
    if (analysis_is_tran_dynamic()) {
      ckt()->CKTmode |= MODETRAN;
    }else if (analysis_is_tran_restore()) {
      ckt()->CKTmode |= MODETRAN;
    }else if (analysis_is_tran_static()) {
      //ckt()->CKTmode |= MODETRANOP;
      ckt()->CKTmode |= MODETRAN;
    }
  }else if (command_is_dc()) {
    ckt()->CKTmode |= MODEDCTRANCURVE;
  }else if (command_is_op()) {
    ckt()->CKTmode |= MODEDCOP;
  }else if (command_is_ac()) {
    ckt()->CKTmode |= MODEAC;
  }else{
    // probably setting probes
    assert(!analysis_is_tran());
    assert(ckt()->CKTmode == 0);
  }
  ckt()->CKTabstol = OPT::abstol;
  ckt()->CKTreltol = OPT::reltol;
  ckt()->CKTvoltTol = OPT::vntol;
  ckt()->CKTgmin = OPT::gmin;
  assert_ckt_up_to_date(ckt());
}

#define assert_ckt_localized(ckt) {				\
    assert_ckt_up_to_date(ckt);					\
    assert((ckt)->CKTstat);					\
    DEV_SPICE* d = reinterpret_cast<DEV_SPICE*>((ckt)->CKTstat);\
    assert(d);							\
    assert(dynamic_cast<DEV_SPICE*>(d));			\
    assert((ckt)->CKTtime == d->_time[0]);			\
    assert((ckt)->CKTdelta == d->_dt);				\
    assert((ckt)->CKTmaxOrder == OPT::_keep_time_steps - 1);	\
    if (d->_dt == 0) {						\
      assert((ckt)->CKTag[0] == 0);				\
      assert((ckt)->CKTintegrateMethod == TRAPEZOIDAL);		\
      assert((ckt)->CKTorder == 1);				\
    }else if (d->_time[1] != 0 && d->_method_a == mTRAP) {	\
      assert(conchk((ckt)->CKTag[0], 2 / d->_dt));		\
      assert((ckt)->CKTintegrateMethod == TRAPEZOIDAL);		\
      assert((ckt)->CKTorder == 2);				\
    }else{							\
      assert(conchk((ckt)->CKTag[0], 1 / d->_dt));		\
      assert((ckt)->CKTintegrateMethod == TRAPEZOIDAL);		\
      assert((ckt)->CKTorder == 1);				\
    }								\
    assert((ckt)->CKTag[0] == (ckt)->CKTag[0]);			\
    assert((ckt)->CKTrhs == d->_i0);				\
    assert((ckt)->CKTrhsOld == d->_v1);				\
    assert((ckt)->CKTirhs == d->_i1);				\
    assert((ckt)->CKTomega == SIM::jomega.imag());		\
    assert((ckt)->CKTtimePoints == d->_time);			\
  }

void DEV_SPICE::localize_ckt()
{
  assert(OPT::_keep_time_steps < 8);
  assert_ckt_up_to_date(ckt());
  ckt()->CKTstat = reinterpret_cast<STATistics*>(this);
  ckt()->CKTtime = _time[0];
  ckt()->CKTdeltaOld[0] = ckt()->CKTdelta = _dt;
  ckt()->CKTdeltaOld[1] = _time[1] - _time[2];
  ckt()->CKTmaxOrder = OPT::_keep_time_steps - 1;
  if (_dt == 0) {
    ckt()->CKTag[1] = ckt()->CKTag[0] = 0;
    ckt()->CKTintegrateMethod = TRAPEZOIDAL; //GEAR;
    ckt()->CKTorder = 1;
  }else if (_time[1] != 0 && _method_a == mTRAP) {
    ckt()->CKTag[0] = 2 / _dt;
    ckt()->CKTag[1] = 1;
    ckt()->CKTintegrateMethod = TRAPEZOIDAL;
    ckt()->CKTorder = 2;
  }else{
    ckt()->CKTag[0] =  1 / _dt;
    ckt()->CKTag[1] = -1 / _dt;
    ckt()->CKTintegrateMethod = TRAPEZOIDAL; //GEAR;
    ckt()->CKTorder = 1;
  }
  ckt()->CKTrhs = _i0;
  ckt()->CKTrhsOld = _v1;
  ckt()->CKTirhs = _i1;

  ckt()->CKTmode &= ~INITF;
  if (SIM::uic_now()) {
    ckt()->CKTmode |= MODEINITTRAN;
    ckt()->CKTmode |= MODEUIC;
  }else if (SIM::initial_step()) {
    ckt()->CKTmode |= MODEINITJCT;
  }else{
    ckt()->CKTmode |= MODEINITFLOAT;
  }

  ckt()->CKTomega = SIM::jomega.imag();
  ckt()->CKTtimePoints = _time;
  assert_ckt_localized(ckt());
}

#define assert_model_raw() {			\
    assert(_gen_model_raw.GENmodType == 0);	\
    assert(_gen_model_raw.GENnextModel == NULL);\
    assert(_gen_model_raw.GENinstances == NULL);\
  }
#define assert_model_setup() {			\
    assert(_model_setup);			\
    assert(_model_setup.GENmodType() == 0);	\
    assert(_model_setup.GENnextModel() == NULL);\
    assert(_model_setup.GENinstances());	\
    assert(_model_setup.GENmodName());		\
  }
#define assert_instance() {			\
    assert(_geninst.GENmodPtr == NULL);	\
    assert(_geninst.GENnextInstance == NULL);\
    assert(_geninst.GENname == NULL);	\
  }
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
struct IFVA {
  IFvalue* _v;
  int _type;
  IFVA(IFvalue* v, int t) :_v(v), _type(t) {assert(v);}
  std::string string()const {
    assert(_v);
    switch (_type & 0xff) {
    case IF_FLAG:
    case IF_INTEGER:
      return to_string(_v->iValue);
    case IF_REAL:
      return to_string(_v->rValue);
    case IF_COMPLEX:untested();
      return ("(" + to_string(_v->cValue.real) + "," + to_string(_v->cValue.imag) + ")");
    case IF_STRING:
      if (_v->sValue) {
	return _v->sValue;
      }else{
	return  "====NULL====";
      }
    default:
      unreachable();
      return "unreachable???";
    }
  }
  void operator=(const std::string& s)
  {
    CS cmd(CS::_STRING, s);
    assert(_v);
    int datatype = _type;
    if (datatype & IF_SET) {
      if (datatype & IF_VECTOR) {
	incomplete();
      }
      switch (datatype & 0xff) {
      case IF_FLAG:	_v->iValue = true;	break;
      case IF_INTEGER:	cmd >> _v->iValue;	break;
      case IF_REAL:	cmd >> _v->rValue;	break;
      case IF_COMPLEX:
	//cmd >> _v->cValue;
	incomplete();
	break;
      case IF_NODE:	incomplete();		break;
      case IF_STRING:
	{
	  //assert(!(_v->sValue));
	  //BUG//memory leak -- this is never deleted
	  _v->sValue = new char[s.length()+1];
	  strcpy(_v->sValue, s.c_str());
	  break;
	}
      case IF_INSTANCE:	incomplete();		break;
      case IF_PARSETREE: incomplete();		break;
      default:		unreachable();		break;
      }
    }else{
    }
  }
};
/*--------------------------------------------------------------------------*/
bool get(CS& cmd, const std::string& key, IFVA* v)
{
  std::string real_key(key);
  if (cmd.umatch(real_key + ' ')) {
    cmd.skip1('=');

    assert(v);
    assert(v->_v);
    int datatype = v->_type;
    if (datatype & IF_SET) {
      if (datatype & IF_VECTOR) {
	incomplete();
      }
      switch (datatype & 0xff) {
      case IF_FLAG:
	v->_v->iValue = true;
	return true;
      case IF_INTEGER:
	cmd >> v->_v->iValue;
	return true;
      case IF_REAL:
	cmd >> v->_v->rValue;
	return true;
      case IF_COMPLEX:
	//cmd >> v->_v->cValue;
	incomplete();
	return false;
      case IF_NODE:
	incomplete();
	return false;
      case IF_STRING:
	{untested();
	  assert(!(v->_v->sValue));
	  std::string s;
	  cmd >> s;
	  //BUG//memory leak -- this is never deleted
	  v->_v->sValue = new char[s.length()+1];
	  strcpy(v->_v->sValue, s.c_str());
	  return true;
	}
      case IF_INSTANCE:
	incomplete();
	return false;
      case IF_PARSETREE:
	incomplete();
	return false;
      default:
	unreachable();
	return false;
      }
    }else{
      return false;
    }
  }else{
    return false;
  }
}
/*--------------------------------------------------------------------------*/
OMSTREAM& operator<<(OMSTREAM& o, const IFVA& v)
{
  assert(v._v);
  switch (v._type & 0xff) {
  case IF_FLAG:
  case IF_INTEGER:
    o << v._v->iValue;
    break;
  case IF_REAL:
    o << v._v->rValue;
    break;
  case IF_COMPLEX:untested();
    o << "(" << v._v->cValue.real << "," << v._v->cValue.imag << ")";
    break;
  case IF_STRING:
    if (v._v->sValue) {
      o << v._v->sValue;
    }else{
      o << "====NULL====";
    }
    break;
  default:untested();
    unreachable();
    break;
  }
  return o;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
MODEL_SPICE::MODEL_SPICE(const DEV_SPICE* p) 
  :MODEL_CARD(p),
   _model_raw(),
   _model_setup(),
   _key(),
   _level()
{
  assert_model_raw();
  std::fill_n(&_model_raw_space, sizeof(MODEL), '\0');
  assert_model_raw();
  try {
    elaborate();
  }catch (Exception&) {untested();
    error(bDANGER, "\nerror in spice model\n");
  }
}
/*--------------------------------------------------------------------------*/
MODEL_SPICE::MODEL_SPICE(const MODEL_SPICE& p)
  :MODEL_CARD(p),
   _model_raw(p._model_raw),
   _model_setup(),
   _key(),
   _level(p._level)
{
  assert_model_raw();
}
/*--------------------------------------------------------------------------*/
MODEL_SPICE::~MODEL_SPICE()
{
  --_count;
  if (_model_setup) {
    delete _model_setup._m;
  }else{
  }
}
/*--------------------------------------------------------------------------*/
void MODEL_SPICE::elaborate()
{
  MODEL_CARD::elaborate();

  init_ckt();
  if (info.DEVsetup) {
    assert_model_raw();

    assert(!_model_setup);
    _model_setup._m = new MODEL(_model_raw);
    assert(_model_setup);
    assert_model_raw();
    assert(_model_setup.GENinstances() == NULL);
    
    int ok = info.DEVsetup(NULL, _model_setup, ckt(), NULL);
    assert(ok == OK);
  }else{
  }
}
/*--------------------------------------------------------------------------*/
void MODEL_SPICE::set_dev_type(const std::string& new_type)
{
  assert_model_raw();

  //_gen_model_raw.set_mod_name(short_label());
  std::string s = short_label();
  char* p = new char[s.length()+1]; //BUG//memory leak
  s.copy(p, std::string::npos);
  p[s.length()] = '\0';
  _gen_model_raw.GENmodName = p;

  _key = new_type;
  set_param_by_name(_key, "1");
}
/*--------------------------------------------------------------------------*/
bool MODEL_SPICE::param_is_printable(int i)const
{
  assert_model_raw();
  assert(!_model_setup);
  assert(info.DEVpublic.numModelParms);
  assert(info.DEVpublic.modelParms);
  assert(info.DEVmodAsk);

  if (i < (*(info.DEVpublic.numModelParms))) {
    IFparm Parms = info.DEVpublic.modelParms[i];
    IFvalue value;
    int datatype = Parms.dataType;
    return (!(datatype & IF_REDUNDANT) 
	    && datatype & IF_SET
	    && datatype & IF_ASK
	    && (info.DEVmodAsk(ckt(), &_gen_model_raw, Parms.id, &value) == OK)
	    );
  }else{
    i -= *(info.DEVpublic.numModelParms);
    switch (i) {
    case 0: return true;
    default: return MODEL_CARD::param_is_printable(i-1);
    }
  }
}
/*--------------------------------------------------------------------------*/
std::string MODEL_SPICE::param_name(int i)const
{
  assert_model_raw();
  assert(!_model_setup);
  assert(info.DEVpublic.numModelParms);
  assert(info.DEVpublic.modelParms);
  assert(info.DEVmodAsk);

  if (i < (*(info.DEVpublic.numModelParms))) {
    static char buffer[20];
    IFparm Parms = info.DEVpublic.modelParms[i];
    // strcpy with an extra \0
    int j = 0;
    for (   ; Parms.keyword[j]; ++j) {
      buffer[j] = Parms.keyword[j];
    }
    buffer[j] = '\0';
    buffer[j+1] = '\0';
    return buffer;
  }else{
    i -= *(info.DEVpublic.numModelParms);
    switch (i) {
    case 0: return "level\0";
    default: return MODEL_CARD::param_name(i-1);
    }
  }
}
/*--------------------------------------------------------------------------*/
std::string MODEL_SPICE::param_value(int i)const
{
  assert_model_raw();
  assert(!_model_setup);
  assert(info.DEVpublic.numModelParms);
  assert(info.DEVpublic.modelParms);
  assert(info.DEVmodAsk);

  if (i < (*(info.DEVpublic.numModelParms))) {
    IFparm Parms = info.DEVpublic.modelParms[i];
    IFvalue value;
    int ok = info.DEVmodAsk(ckt(), &_gen_model_raw, Parms.id, &value);
    if (ok == OK) {
      return IFVA(&value,Parms.dataType).string();
    }else{untested();
      return "";
    }
  }else{
    i -= *(info.DEVpublic.numModelParms);
    switch (i) {
    case 0: return _level;
    default: return MODEL_CARD::param_value(i-1);
    }
  }
}
/*--------------------------------------------------------------------------*/
void MODEL_SPICE::set_param_by_index(int i, std::string& new_value, int offset)
{
  assert_model_raw();
  assert(!_model_setup);
  assert(info.DEVpublic.numModelParms);
  assert(info.DEVpublic.modelParms);
  assert(info.DEVmodParam);

  if (i < (*(info.DEVpublic.numModelParms))) {
    IFparm Parms = info.DEVpublic.modelParms[i];
    IFvalue value;
    IFVA v(&value, Parms.dataType);
    v = new_value;
    int ok = info.DEVmodParam(Parms.id, &value, &_gen_model_raw);
    assert(ok == OK);
  }else{
    i -= *(info.DEVpublic.numModelParms);
    offset += *(info.DEVpublic.numModelParms);
    switch (i) {
    case 0: _level = new_value; break;
    default: CARD::set_param_by_index(i-1, new_value, offset+1); break;
    }
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
DEV_SPICE::DEV_SPICE()
  :STORAGE(),
   _inst(),
   _modelname(""),
   _model_setup(),
   _nodes(),
   _matrix(),
   _matrix_core(),
   _i0(),
   _i1(),
   _v1(),
   _state0(NULL),
   _state1(NULL),
   _state2(NULL),
   _state3(NULL),
   _state4(NULL),
   _num_states(0)
{
  assert(OPT::_keep_time_steps == 5);
  assert_instance();
  std::fill_n(&_inst_space, sizeof(INSTANCE), '\0');
  assert_instance();

  {
    int* node = spice_nodes();
    for (int ii = 0; ii < matrix_nodes(); ++ii) {
      node[ii] = SPICE_INVALID_NODE;
    }
  }
  _n = _nodes;
  for (int ii = 0; ii < matrix_nodes(); ++ii) {
    assert(!(_n[ii].n_()));
  }

  for (int ii = 0; ii < matrix_nodes()+OFFSET; ++ii) {
    _matrix[ii] = _matrix_core[ii];
    assert(_matrix[ii]);
  }

  ++_count;
}
/*--------------------------------------------------------------------------*/
DEV_SPICE::DEV_SPICE(const DEV_SPICE& p)
  :STORAGE(p),
   _inst(p._inst),
   _modelname(p._modelname),
   _model_setup(),
   _nodes(),
   _matrix(),
   _matrix_core(),
   _i0(),
   _i1(),
   _v1(),
   _state0(NULL),
   _state1(NULL),
   _state2(NULL),
   _state3(NULL),
   _state4(NULL),
   _num_states(p._num_states)
{
  assert(OPT::_keep_time_steps == 5);
  assert_instance();

  {
    int* node = spice_nodes();
    for (int ii = 0; ii < matrix_nodes(); ++ii) {
      assert(node[ii] == SPICE_INVALID_NODE);
    }
  }
  _n = _nodes;
  for (int ii = 0; ii < matrix_nodes(); ++ii) {
    _n[ii] = p._n[ii];
  }

  for (int ii = 0; ii < matrix_nodes()+OFFSET; ++ii) {
    _matrix[ii] = _matrix_core[ii];
    assert(_matrix[ii]);
  }

  ++_count;
}
/*--------------------------------------------------------------------------*/
DEV_SPICE::~DEV_SPICE()
{
  assert_instance();

  --_count;
  
  if (_state0) {
    // regular instances
    assert(OPT::_keep_time_steps == 5);
    assert(_state0);
    assert(_state1);
    assert(_state2);
    assert(_state3);
    assert(_state4);
    assert(_model_setup);
    
    assert(OPT::_keep_time_steps == 5);
    delete [] _state0;
    delete [] _state1;
    delete [] _state2;
    delete [] _state3;
    delete [] _state4;
    delete _model_setup._m;
  }else{
    // prototype
    assert(OPT::_keep_time_steps == 5);
    assert(!_state0);
    assert(!_state1);
    assert(!_state2);
    assert(!_state3);
    assert(!_state4);
    assert(!_model_setup);
  }
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::set_dev_type(const std::string& new_type)
{
  _modelname = new_type;
}
/*--------------------------------------------------------------------------*/
bool DEV_SPICE::param_is_printable(int i)const
{
  assert_instance();
  assert(info.DEVpublic.numInstanceParms);
  assert(info.DEVpublic.instanceParms);
  assert(info.DEVask);

  int num_params = *(info.DEVpublic.numInstanceParms);
  if (i < num_params) {
    IFparm Parms = info.DEVpublic.instanceParms[i];
    IFvalue Value;
    int datatype = Parms.dataType;
    return (!(datatype & IF_REDUNDANT) 
	    && datatype & IF_SET
	    && datatype & IF_ASK
	    && (info.DEVask(ckt(), &_geninst, Parms.id, &Value, NULL) == OK)
	    );
  }else{
    return STORAGE::param_is_printable(i-num_params);
  }
}
/*--------------------------------------------------------------------------*/
std::string DEV_SPICE::param_name(int i)const
{
  assert_instance();
  assert(info.DEVpublic.numInstanceParms);
  assert(info.DEVpublic.instanceParms);
  assert(info.DEVask);

  int num_params = *(info.DEVpublic.numInstanceParms);
  if (i < num_params) {
    static char buffer[20];
    IFparm Parms = info.DEVpublic.instanceParms[i];
    // strcpy with an extra \0
    int j = 0;
    for (   ; Parms.keyword[j]; ++j) {
      buffer[j] = Parms.keyword[j];
    }
    buffer[j] = '\0';
    buffer[j+1] = '\0';
    return buffer;
  }else{
    return STORAGE::param_name(i-num_params);
  }
}
/*--------------------------------------------------------------------------*/
std::string DEV_SPICE::param_value(int i)const
{
  assert_instance();
  assert(info.DEVpublic.numInstanceParms);
  assert(info.DEVpublic.instanceParms);
  assert(info.DEVask);

  int num_params = *(info.DEVpublic.numInstanceParms);
  if (i < num_params) {
    IFparm Parms = info.DEVpublic.instanceParms[i];
    IFvalue Value;
    int ok = info.DEVask(ckt(), &_geninst, Parms.id, &Value, NULL);
    if (ok == OK) {
      return IFVA(&Value,Parms.dataType).string();
    }else{untested();
      return "";
    }
  }else{
    return STORAGE::param_value(i-num_params);
  }
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::set_param_by_index(int i, std::string& new_value, int offset)
{
  assert_instance();
  assert(info.DEVpublic.numInstanceParms);
  assert(info.DEVpublic.instanceParms);
  assert(info.DEVparam);

  int num_params = *(info.DEVpublic.numInstanceParms);
  if (i < num_params) {
    IFparm Parms = info.DEVpublic.instanceParms[i];
    IFvalue Value;
    IFVA v(&Value, Parms.dataType);
    v = new_value;
#ifdef JSPICE3
    int ok = info.DEVparam(ckt(), Parms.id, &Value, &_geninst, NULL);
#else
    int ok = info.DEVparam(Parms.id, &Value, &_geninst, NULL);
#endif
    assert(ok == OK);
  }else{
    STORAGE::set_param_by_index(i-num_params, new_value, offset+num_params);
  }
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::elaborate()
{
  STORAGE::elaborate();

  init_ckt();
  assert_instance();
  assert(info.DEVsetup);

  // attach model
  const MODEL_SPICE* m = dynamic_cast<const MODEL_SPICE*>(find_model(_modelname));
  if (!m) {untested();
    throw Exception_Model_Type_Mismatch(long_label(), _modelname, DEVICE_TYPE);
  }else{
  }

  //-------- allocate matrix
  for (int ii = 0; ii < matrix_nodes()+OFFSET; ++ii) {
    assert(_matrix[ii]);
    for (int jj = 0; jj < matrix_nodes()+OFFSET; ++jj) {
      _matrix[ii][jj] = 0;
    }
  }
  
  { //-------- fix up external nodes
    int* node = spice_nodes();
    for (int ii = 0; ii < net_nodes(); ++ii) {
      node[ii] = ii+OFFSET;
    }
#if (UNCONNECTED_NODES == GROUND)
    for (int ii = net_nodes(); ii < max_nodes(); ++ii) {
      node[ii] = ii+OFFSET;
    }
    ckt()->CKTmaxEqNum = max_nodes();
#elif (UNCONNECTED_NODES == FLOAT)
    for (int ii = net_nodes(); ii < max_nodes(); ++ii) {
      node[ii] = SPICE_UNCONNECTED_NODE;
    }
    ckt()->CKTmaxEqNum = max_nodes();
#else
    // UNCONNECTED_NODES == DISALLOW
    assert(min_nodes() == max_nodes());
    assert(net_nodes() == max_nodes());
    ckt()->CKTmaxEqNum = net_nodes();
#endif
  }
  
  { //------- fill in defaults, set up matrix, map nodes
    // This kluge is necessary because the Spice setup function
    // does models and instances together, and will not setup 
    // an instance without doing setup on the model also.
    // It is not permissible to setup a model more than once,
    // so it is necessary to make a new one every time, 
    // just so it can be setup, even if it is never used.

    MODEL_PTR model_copy = new MODEL(m->_model_raw);

    SMPmatrix* matrix = reinterpret_cast<SMPmatrix*>(_matrix);
    _num_states = 0;
    model_copy.GENinstances() = &_geninst;
    //-------------
    int ok = info.DEVsetup(matrix, model_copy, ckt(), &_num_states);
    //-------------
    assert(ok == OK);
    if (_model_setup) {
      delete _model_setup._m;
    }else{
    }
    _model_setup = model_copy;
  }
  { //-------- fix up internal nodes
    int* node = spice_nodes(); // treat as array
    char fake_name[] = "a";

#if (UNCONNECTED_NODES == GROUND)
    for (int ii = net_nodes(); ii < max_nodes(); ++ii) {
      _n[ii].set_to_ground(this);
    }
    int start_internal = max_nodes();
#else
    // UNCONNECTED_NODES == DISALLOW || UNCONNECTED_NODES == FLOAT
    int start_internal = net_nodes();
#endif
    for (int ii = start_internal; ii < matrix_nodes(); ++ii) {
      if (node[ii] >= start_internal+OFFSET) {
	// real internal node
	_n[ii].new_model_node(long_label() + '.' + fake_name, this);
	assert(_n[ii].n_());
      }else if (node[ii] >= 0+OFFSET) {
	// collapsed to an external node
	_n[ii] = _n[node[ii]-OFFSET];
	assert(_n[ii].n_());
      }else{
	// not assigned
	assert(!_n[ii].n_());
      }
      ++(*fake_name);
    }

    for (int ii = 0; ii < matrix_nodes(); ++ii) {
      trace2((_n[ii].n_()) ? (_n[ii].n_()->short_label().c_str()) : ("NULL"),
	     ii, node[ii]);
    }

    // This could be one loop, but doing it this way gives more info.
    for (int ii = 0; ii < min_nodes(); ++ii) {
      assert(_n[ii].n_());
    }
    for (int ii = min_nodes(); ii < net_nodes(); ++ii) {
      assert(_n[ii].n_());
    }
    for (int ii = net_nodes(); ii < max_nodes(); ++ii) {
      //assert(_n[ii].n_());
    }
    for (int ii = max_nodes(); ii < matrix_nodes(); ++ii) {
      assert(_n[ii].n_() || !node[ii]);
    }
  }

  { //-------- allocate vectors
    if (!_state0) {
      assert(OPT::_keep_time_steps == 5);
      assert(!_state1);
      assert(!_state2);
      assert(!_state3);
      assert(!_state4);
      _state0 = new double[_num_states];
      _state1 = new double[_num_states];
      _state2 = new double[_num_states];
      _state3 = new double[_num_states];
      _state4 = new double[_num_states];
    }else{
      assert(OPT::_keep_time_steps == 5);
      assert(_state0);
      assert(_state1);
      assert(_state2);
      assert(_state3);
      assert(_state4);
    }
  }

  std::fill_n(_i1, matrix_nodes()+OFFSET, 0);
  std::fill_n(_v1, matrix_nodes()+OFFSET, 0);
  assert(OPT::_keep_time_steps == 5);
  std::fill_n(_state0, _num_states, 0);
  std::fill_n(_state1, _num_states, 0);
  std::fill_n(_state2, _num_states, 0);
  std::fill_n(_state3, _num_states, 0);
  std::fill_n(_state4, _num_states, 0);
  
  assert(!is_constant());
  assert(m->_gen_model_raw.GENinstances == NULL);
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::precalc()
{
  // ELEMENT::precalc(); .. don't call .. more analysis needed
  update_ckt();

  assert_instance();
  assert_model_setup();
  assert(info.DEVtemperature);
  assert(_model_setup.GENinstances() == &_geninst);

  //-----
  info.DEVtemperature(_model_setup, ckt());
  //-----
  //_model_setup->GENinstances = NULL;
  set_converged();
  assert(!is_constant());
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::tr_restore()
{
  STORAGE::tr_restore();
  update_ckt();
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::tr_advance()
{
  STORAGE::tr_advance();
  update_ckt();
  
  assert(OPT::_keep_time_steps == 5);
  double* t = _state4;
  _state4 = _state3;
  _state3 = _state2;
  _state2 = _state1;
  _state1 = _state0;
  _state0 = t;
  std::fill_n(_state0, _num_states, 0);
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::tr_regress()
{untested();
  ELEMENT::tr_regress();
  update_ckt();
}
/*--------------------------------------------------------------------------*/
bool DEV_SPICE::do_tr()
{
  localize_ckt();

  assert_instance();
  assert_model_setup();
  assert(info.DEVload);
  assert(OPT::_keep_time_steps == 5);
  assert(_state0);
  assert(_state1);
  assert(_state2);
  assert(_state3);
  assert(_state4);
  assert(_num_states >= 0);
  
  // copy in
  assert(_model_setup.GENinstances() == &_geninst);
  ckt()->CKTdelta = _dt;

  std::fill_n(_v1, matrix_nodes()+OFFSET, 0);
  int* node = spice_nodes();
  for (int ii = 0; ii < matrix_nodes(); ++ii) {
    if (node[ii] != SPICE_INVALID_NODE) {
      _v1[node[ii]] = _n[ii].v0();
    }else{
    }
  }

  // copy in copy out
  ckt()->CKTtroubleElt = NULL;
  ckt()->CKTnoncon = 0;
  assert(OPT::_keep_time_steps == 5);
  ckt()->CKTstate0 = _state0;
  ckt()->CKTstate1 = _state1; // copy out is dv kluge
  ckt()->CKTstate2 = _state2;
  ckt()->CKTstate3 = _state3;
  ckt()->CKTstate4 = _state4;

  std::fill_n(_i0, matrix_nodes()+OFFSET, 0);
  for (int ii = 0; ii < matrix_nodes()+OFFSET; ++ii) {
    for (int jj = 0; jj < matrix_nodes()+OFFSET; ++jj) {
      _matrix[ii][jj].real() = 0;
    }
  }

  // do the work -- it might also do convergence checking, might not
  //-----
  info.DEVload(_model_setup, ckt());
  //-----

  for (int ii = 0; ii < matrix_nodes()+OFFSET; ++ii) {
    trace3("conchk", ii, _i0[ii], _i1[ii]);
  }

  // convergence check -- gnucap method
  set_converged(ckt()->CKTnoncon == 0);
  for (int ii = 0; converged() && ii < matrix_nodes()+OFFSET; ++ii) {
    set_converged(conchk(_i0[ii], _i1[ii]));
  }
  for (int ii = 0; converged() && ii < matrix_nodes()+OFFSET; ++ii) {
    for (int jj = 0; converged() && jj < matrix_nodes()+OFFSET; ++jj) {
      set_converged(conchk(_matrix[ii][jj].real(), _matrix[ii][jj].imag()));
    }
  }

  // convergence check -- Spice method 
  // not sure if it is worth the effort
  if (converged() && info.DEVconvTest) {
    ckt()->CKTnoncon = 0;
    ckt()->CKTrhs = _v1;
    info.DEVconvTest(_model_setup, ckt());
    set_converged(ckt()->CKTnoncon == 0);
  }else{
    // either no separate test or already failed
  }

  q_load();
  return converged();
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::tr_load()
{
#ifndef NDEBUG
  if (_loaditer == iteration_tag()) {
    error(bDANGER, long_label() + " internal error: double load\n");
  }
  _loaditer = iteration_tag();
#endif

  int ihit[MATRIX_NODES+OFFSET];
  int jhit[MATRIX_NODES+OFFSET];

  std::fill_n(ihit, matrix_nodes()+OFFSET, 0);
  std::fill_n(jhit, matrix_nodes()+OFFSET, 0);

  int* node = spice_nodes();
  for (int ii = 0; ii < matrix_nodes(); ++ii) {
    int ni = node[ii];
    if (ni && !ihit[ni]) {
      ihit[ni] = 1;
      int nii = ni-OFFSET;
      trace4("", ii, ni, _i0[ni], _i1[ni]);
      tr_load_source_point(_n[ii], &(_i0[ni]), &(_i1[ni]));
      for (int jj = 0; jj < matrix_nodes(); ++jj) {
	int nj = node[jj];
	if (nj && jhit[nj] != ni) {
	  jhit[nj] = ni;
	  int njj = nj-OFFSET;
	  trace2("", jj, nj);
	  trace2("", _matrix[nii][njj].real(), _matrix[nii][njj].imag());
	  tr_load_point(_n[ii], _n[jj],
			&(_matrix[nii][njj].real()), &(_matrix[nii][njj].imag()));
	}else{
	  trace2("skip", jj, nj);
	}
      }
    }else{
      trace2("=========skip", ii, ni);
    }
  }
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::tr_unload()
{untested();incomplete();

  for (int ii = 0; ii < matrix_nodes(); ++ii) {untested();
    for (int jj = 0; jj < matrix_nodes(); ++jj) {untested();
      _matrix[ii][jj].real() = 0;
    }
  }
  mark_inc_mode_bad();
  tr_load();
}
/*--------------------------------------------------------------------------*/
TIME_PAIR DEV_SPICE::tr_review()
{
  // not calling STORAGE::tr_review();
  _time_by.reset();

  localize_ckt();
  assert_instance();
  assert_model_setup();

  if (info.DEVtrunc) {
    // copy in
    ckt()->CKTdelta = _dt;
    assert(OPT::_keep_time_steps == 5);
    ckt()->CKTstate0 = _state0;
    ckt()->CKTstate1 = _state1;
    ckt()->CKTstate2 = _state2;
    ckt()->CKTstate3 = _state3;
    ckt()->CKTstate4 = _state4;
    
    // parameter passing
    assert(_model_setup.GENinstances() == &_geninst);
    ckt()->CKTtroubleElt = NULL;
    double timestep = NEVER;
    
    //-----
    info.DEVtrunc(_model_setup, ckt(), &timestep);
    //-----
    
    // clean-up
    //_model_setup->GENinstances = NULL;
    
    _time_by._error_estimate = tr_review_check_and_convert(timestep);
  }else{
  }
  return _time_by;
}
/*--------------------------------------------------------------------------*/
double DEV_SPICE::tr_probe_num(const std::string& x)const
{
  assert_ckt_up_to_date(ckt());
  assert_instance();
  assert(info.DEVpublic.numInstanceParms);
  assert(info.DEVpublic.instanceParms);
  assert(info.DEVask);

  // all of the "states" in state array
  for (int ii=0; ii<_num_states && state_names[ii]!=""; ++ii) {
    if (Umatch(x, state_names[ii] + ' ')) {
      return _state0[ii];
    }else{
    }
  }

  // data that Spice has, through "ask"
  assert(OPT::_keep_time_steps == 5);
  ckt()->CKTstate0 = _state0;
  ckt()->CKTstate1 = _state1;
  ckt()->CKTstate2 = _state2;
  ckt()->CKTstate3 = _state3;
  ckt()->CKTstate4 = _state4;
  for (int ii=0; ii<(*(info.DEVpublic.numInstanceParms)); ++ii) {
    IFparm Parms = info.DEVpublic.instanceParms[ii];
    int datatype = Parms.dataType;
    if (datatype & IF_ASK && Umatch(x, std::string(Parms.keyword) + ' ')) {
      IFvalue v;
      int ok = info.DEVask(ckt(), &_geninst, Parms.id, &v, NULL);
      if (ok == OK) {
	switch (datatype & 0xff) {
	case IF_FLAG:
	case IF_INTEGER:
	  return v.iValue;
	case IF_REAL:
	  return v.rValue;
	case IF_COMPLEX:
	case IF_STRING:
	default:
	  unreachable();
	  break;
	}
      }else{untested();
      }
    }else{
    }
  }
  return STORAGE::tr_probe_num(x);
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::do_ac()
{
  localize_ckt();
  assert_instance();
  assert_model_setup();
  assert(info.DEVacLoad);
  assert(OPT::_keep_time_steps == 5);
  assert(_state0);
  assert(_state1);
  assert(_state2);
  assert(_state3);
  assert(_state4);
  assert(_num_states >= 0);
  
  // copy in
  std::fill_n(_v1, matrix_nodes()+OFFSET, 0);
  int* node = spice_nodes();
  for (int ii = 0; ii < matrix_nodes(); ++ii) {itested();
    //BUG// complex???????
    _v1[node[ii]] = _n[ii].v0();
  }

  // copy in copy out
  assert(OPT::_keep_time_steps == 5);
  ckt()->CKTstate0 = _state0;
  ckt()->CKTstate1 = _state1; // copy out is dv kluge
  ckt()->CKTstate2 = _state2;
  ckt()->CKTstate3 = _state3;
  ckt()->CKTstate4 = _state4;

  // clear for copy out
  ckt()->CKTtroubleElt = NULL;

  std::fill_n(_i0, matrix_nodes()+OFFSET, 0);
  std::fill_n(_i1, matrix_nodes()+OFFSET, 0);

  for (int ii = 0; ii < matrix_nodes()+OFFSET; ++ii) {
    for (int jj = 0; jj < matrix_nodes()+OFFSET; ++jj) {
      _matrix[ii][jj] = 0;
    }
  }
  
  // parameter passing
  assert(_model_setup.GENinstances() == &_geninst);

  //-----
  info.DEVacLoad(_model_setup, ckt());
  //-----

  ac_load();
}
/*--------------------------------------------------------------------------*/
void DEV_SPICE::ac_load()
{
  assert_ckt_up_to_date(ckt());

  int* node = spice_nodes();
  for (int ii = 0; ii < matrix_nodes(); ++ii) {
    ac_load_source_point(_n[ii], COMPLEX(_i0[node[ii]], _i1[node[ii]]));
    for (int jj = 0; jj < matrix_nodes(); ++jj) {
      ac_load_point(_n[ii], _n[jj], _matrix[ii][jj]);
    }
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
extern "C" {
  // needed to satisfy references.  Supposedly unreachable.  Stubs.
  char *errMsg = NULL;
  char *errRtn = NULL;
#ifndef NGSPICE_17
  char* tmalloc(int) {incomplete(); return NULL;} //on errors, always a leak
  char* trealloc(char*, int) {incomplete(); return NULL;} //DEVnoise
#endif
 
  IFuid CKTnodName(CKTcircuit*,int) {incomplete();return IFuid();} //DEVsenPrint

  double D1i2F1(double, double, double) {incomplete(); return NOT_VALID;} //DEVdisto
  double D1i3F1(double, double, double, double, double, double) {incomplete(); return NOT_VALID;}
  double D1iF12(double, double, double, double, double) {incomplete(); return NOT_VALID;}
  double D1i2F12(double, double, double, double, double, double, double, double, double,
		 double) {incomplete(); return NOT_VALID;}
  double D1n2F1(double, double, double) {incomplete(); return NOT_VALID;}
  double D1n3F1(double, double, double, double, double, double) {incomplete(); return NOT_VALID;}
  double D1nF12(double, double, double, double, double) {incomplete(); return NOT_VALID;}
  double D1n2F12(double, double, double, double, double, double, double, double, double,
		 double) {incomplete(); return NOT_VALID;}
  double DFn2F1(double, double, double, double, double, double, double, double, double,
		double, double, double) {incomplete(); return NOT_VALID;}
  double DFi2F1(double, double, double, double, double, double, double, double, double,
		double, double, double) {incomplete(); return NOT_VALID;}
  double DFi3F1(double, double, double, double, double, double, double, double, double,
		double, double, double, double, double, double, double, double, double,
		double, double, double, double, double, double, double, double, double,
		double) {incomplete(); return NOT_VALID;}
  double DFn3F1(double, double, double, double, double, double, double, double, double,
		double, double, double, double, double, double, double, double, double,
		double, double, double, double, double, double, double, double, double,
		double) {incomplete(); return NOT_VALID;}
  double DFnF12(double, double, double, double, double, double, double, double,
		double, double, double, double, double, double, double, double,
		double, double) {incomplete(); return NOT_VALID;}
  double DFiF12(double, double, double, double, double, double, double, double,
		double, double, double, double, double, double, double, double,
		double, double) {incomplete(); return NOT_VALID;}
  struct DpassStr; //DEVdisto
  double DFn2F12(DpassStr*) {incomplete(); return NOT_VALID;}
  double DFi2F12(DpassStr*) {incomplete(); return NOT_VALID;}
  struct Dderivs;
  void AtanDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void CosDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void CubeDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void DivDeriv(Dderivs*, Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void EqualDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void ExpDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void InvDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void MultDeriv(Dderivs*, Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void PlusDeriv(Dderivs*, Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void PowDeriv(Dderivs*, Dderivs*, double) {incomplete();} //DEVdisto
  void SqrtDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void TanDeriv(Dderivs*, Dderivs*) {incomplete();} //DEVdisto
  void TimesDeriv(Dderivs*, Dderivs*, double) {incomplete();} //DEVdisto
#ifdef JSPICE3
  double Nintegrate(double, double, double, GENERIC*) {incomplete(); return NOT_VALID;} //DEVnoise
#else
  double Nintegrate(double, double, double, Ndata*) {incomplete(); return NOT_VALID;} //DEVnoise
#endif
  void NevalSrc(double*, double*, CKTcircuit*, int, int, int, double) {incomplete();} //DEVnoise
  void NevalSrc2(double*, double*, CKTcircuit*, int, int, int, double, double) {incomplete();}
  //------------------------------------------------
  // should be constants, but spice wants them to be variables.
  double CONSTroot2(sqrt(2.));
  double CONSTvt0(P_CELSIUS0*P_K_Q);
  double CONSTKoverQ(P_K_Q);
  double CONSTe(M_E);
  //------------------------------------------------
  // ngspice baggage
  int ARCHme = 0;
  // jspice baggage
  IFsimulator *ft_sim;
  //------------------------------------------------
  //------------------------------------------------
  int IFerror(int spice_code, char* message, IFuid* id) /* output an error or warning message */
  {
    switch (spice_code) {
    case ERR_WARNING:error(bWARNING,message, reinterpret_cast<char*>(id)); break;
    case ERR_FATAL:  error(bDANGER, message, reinterpret_cast<char*>(id)); throw Exception("");
    case ERR_PANIC:  error(bDANGER, message, reinterpret_cast<char*>(id)); throw Exception("");
    case ERR_INFO:   error(bTRACE,  message, reinterpret_cast<char*>(id)); break;
    default:	     error(bDANGER, message, reinterpret_cast<char*>(id)); break;
    }
    return OK;
  }
  void internalerror(char *message)
  {
    error(bDANGER, "internal error: %s\n", message);
  }
    
#ifdef NGSPICE_17
  int CKTinst2Node(void*, void*, int, CKTnode**, IFuid*)
  {
    return 10;
  }
#endif
#ifdef JSPICE3
  static IFfrontEnd fe = {
    NULL, //int ((*IFnewUid)());	/* create a new UID in the circuit */ noise, urcsetup
    NULL, //int ((*IFpauseTest)());	/* should we stop now? */ noisean.c only
    NULL, //double ((*IFseconds)());	/* what time is it? */ bjtdisto unused ifdef only (unused)
    IFerror, //int ((*IFerror)());	/* output an error or warning message */ temp, setup
    NULL,
    NULL,
    NULL,
    NULL,
    NULL 
  };
#else
  static IFfrontEnd fe = {
    NULL, //int ((*IFnewUid)());	/* create a new UID in the circuit */ noise, urcsetup
    NULL, //int ((*IFdelUid)());	/* create a new UID in the circuit */ not used
    NULL, //int ((*IFpauseTest)());	/* should we stop now? */ noisean.c only
    NULL, //double ((*IFseconds)());	/* what time is it? */ bjtdisto unused ifdef only (unused)
    IFerror, //int ((*IFerror)());	/* output an error or warning message */ temp, setup
    NULL, //int ((*OUTpBeginPlot)());	/* start pointwise output plot */ noisean.c only
    NULL, //int ((*OUTpData)());	/* data for pointwise plot */ noisean.c only
    NULL, //int ((*OUTwBeginPlot)());	/* start windowed output plot */ not used
    NULL, //int ((*OUTwReference)());	/* independent vector for windowed plot */ not used
    NULL, //int ((*OUTwData)());	/* data for windowed plot */ not used
    NULL, //int ((*OUTwEnd)());		/* signal end of windows */ not used
    NULL, //int ((*OUTendPlot)());	/* end of plot */ not used
    NULL, //int ((*OUTbeginDomain)());	/* start nested domain */ not used
    NULL, //int ((*OUTendDomain)());	/* end nested domain */ not used
    NULL  //int ((*OUTattributes)());	/* specify output attributes of node */ noisean.c only
  };
#endif
  IFfrontEnd* SPfrontEnd = &fe;
  //------------------------------------------------
  //------------------------------------------------
  int CKTsetBreak(CKTcircuit* ckt, double time)
  {untested();
    if (time < ckt->CKTminBreak) {
      ckt->CKTminBreak = time;
    }else{
    }
    return OK;
  }
  //------------------------------------------------
  void CKTterr(int qcap, CKTcircuit* ckt,double *time_step)
  {
    assert_ckt_localized(ckt);

    std::valarray<FPOLY1> q(OPT::_keep_time_steps);

    for (int ii = 0; ii < OPT::_keep_time_steps; ++ii) {
      assert(ckt->CKTstates[ii]);
      q[ii].x  = NOT_VALID;
      q[ii].f0 = ckt->CKTstates[ii][qcap];
      q[ii].f1 = NOT_VALID;
    }

    DEV_SPICE* d = reinterpret_cast<DEV_SPICE*>(ckt->CKTstat);
    assert(d);
    assert(dynamic_cast<DEV_SPICE*>(d));

    *time_step = std::min(d->tr_review_trunc_error(&q[0]), *time_step);
  }
  //------------------------------------------------
  int NIintegrate(CKTcircuit* ckt,double* geq,double* ceq,double cap,int qcap)
  { //-- used by DEVload (not DC)
    assert_ckt_localized(ckt);

    METHOD method;
    if (ckt->CKTorder == 1) {
      method = mEULER;
    }else if (ckt->CKTintegrateMethod == TRAPEZOIDAL) {
      assert(ckt->CKTtimePoints[1] != 0.);
      assert(ckt->CKTorder == 2);
      method = mTRAP;
    }else{
      assert(ckt->CKTintegrateMethod == TRAPEZOIDAL); //GEAR);
      method = mGEAR;
    }

    std::valarray<FPOLY1> q(OPT::_keep_time_steps);
    std::valarray<FPOLY1> i(OPT::_keep_time_steps);

    for (int ii = 0; ii < OPT::_keep_time_steps; ++ii) {
      assert(ckt->CKTstates[ii]);
      q[ii].x  = NOT_VALID;
      q[ii].f0 = ckt->CKTstates[ii][qcap];
      q[ii].f1 = cap;
      trace3("", ii, q[ii].f0, q[ii].f1);
      i[ii].x  = NOT_VALID;
      i[ii].f0 = ckt->CKTstates[ii][qcap+1];
      i[ii].f1 = q[ii].f1 * ckt->CKTag[0];
      trace3("", ii, i[ii].f0, i[ii].f1);
      assert(q[ii].f0 == q[ii].f0);
      assert(q[ii].f1 == q[ii].f1);
      assert(i[ii].f0 == i[ii].f0);
      assert(i[ii].f1 == i[ii].f1);
    }

    i[0] = differentiate(&q[0], &i[0], ckt->CKTtimePoints, method);
    assert(i[0].f0 == i[0].f0);
    assert(i[0].f1 == i[0].f1);
    trace2("", i[0].f0, i[0].f1);

    ckt->CKTstates[0][qcap+1] = i[0].f0;

    assert(ckt->CKTdelta != 0. || (ckt->CKTag[0] == 0. && i[0].f0 == 0.));
    *ceq = i[0].f0 - q[0].f0 * ckt->CKTag[0];
    *geq = i[0].f1;
    assert(*ceq == *ceq);
    assert(*geq == *geq);
    trace2("", *ceq, *geq);
    return OK;
  }
  //------------------------------------------------
  //------------------------------------------------
  int CKTmkVolt(CKTcircuit* ckt, CKTnode** n, IFuid, char*)
  { // get a new node number. -- used by DEVsetup
    assert_ckt_initialized(ckt);
    assert(n);
    static CKTnode n_static;	// always used only on next line
    *n = &n_static;		// so reuse static structure
    (*n)->number = ((ckt->CKTmaxEqNum)++)+OFFSET;
    // local number (- == internal)   only number is used
    return OK;
  }
  int CKTmkCur(CKTcircuit* ckt, CKTnode** n, IFuid i, char* c)
  {
    return CKTmkVolt(ckt, n, i, c);
  }
  //------------------------------------------------  
  int CKTdltNNum(void*,int)
  { // complement to CKTmkVolt. -- used by DEVunsetup
    // deletes what was new in CKTmkVolt
    // Nothing, because of no alloc there.
    return OK;
  }
  //------------------------------------------------  
  //------------------------------------------------  
  double* SMPmakeElt(SMPmatrix* mm, int r, int c)
  { // returns a pointer m[r][c] -- used by DEVsetup
    assert(mm);
    trace2("", r, c);
    if (r == 0 || c == 0) {
      static double trash;
      trash = 0;
      return &trash;
    }else{
      assert(r >= 0+OFFSET);
      assert(r < MATRIX_NODES+OFFSET);
      assert(c >= 0+OFFSET);
      assert(c < MATRIX_NODES+OFFSET);
      COMPLEX** m = reinterpret_cast<COMPLEX**>(mm);
      assert(m);
      assert(m[r-OFFSET]);
      return reinterpret_cast<double*>(&(m[r-OFFSET][c-OFFSET]));
    }
  }
  //------------------------------------------------  
#ifdef JSPICE3
  int IFnewUid(GENERIC*,IFuid*,IFuid,char*,int,GENERIC**) {incomplete(); return 0;}
  int INPpName(char*,IFvalue*,GENERIC*,int,GENERIC*) {incomplete(); return 0;}
  char *INPdevErr(char *) {incomplete(); return NULL;}
  char *INPerror(int) {incomplete(); return NULL;}
  spREAL *spGetElement(char* s, int r, int c) {return SMPmakeElt(s,r,c);}
  char *INPerrCat(char *, char *) {incomplete(); return NULL;}
  int INPgndInsert(GENERIC*,char**,INPtables*,GENERIC**) {incomplete(); return 0;}
  char * INPdevParse(char**,GENERIC*,int,GENERIC*,double*,int*,INPtables*) {incomplete(); return NULL;}
  char *INPgetMod(GENERIC*,char*,INPmodel**,INPtables*) {incomplete(); return NULL;}
  int INPgetTok(char**,char**,int) {incomplete(); return 0;}
  int INPlookMod(char*) {incomplete(); return 0;}
  int INPtermInsert(GENERIC*,char**,INPtables*,GENERIC**) {incomplete(); return 0;}
  int INPinsert(char**,INPtables*) {incomplete(); return 0;}
  char *copy(char*) {incomplete(); return NULL;}
  int NIsum(CKTcircuit*,double*,int) {incomplete(); return 0;}
  double INPevaluate(char**,int*,int) {incomplete(); return NOT_VALID;}
  IFvalue *INPgetValue(GENERIC*,char**,int,INPtables*) {incomplete(); return NULL;}
#endif
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
// Verify that the layout of complex is as Spice assumes.
// This is not guaranteed by the standard, but is believed to always be true.
static struct COMPLEX_TEST {
  COMPLEX_TEST() {
    COMPLEX x;
    COMPLEX* px = &x;
    double* prx = &x.real();
    double* pix = &x.imag();
    assert(reinterpret_cast<void*>(prx) == reinterpret_cast<void*>(px));
    assert(reinterpret_cast<void*>(pix-1) == reinterpret_cast<void*>(px));
  }
  ~COMPLEX_TEST() {
  }  
} complex_test;
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
int MODEL_SPICE::_count = -1;
int DEV_SPICE::_count = -1;

static DEV_SPICE p0;
static DISPATCHER<CARD>::INSTALL
  d0(&device_dispatcher, std::string(SPICE_LETTER) + "|" + DEVICE_TYPE, &p0);

static MODEL_SPICE p1(&p0);
static DISPATCHER<MODEL_CARD>::INSTALL
  d1(&model_dispatcher, MODEL_TYPE, &p1);
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
